@@ -222,29 +222,43 @@ class Connection(object):
         """
         return self._socket is None
 
+    def _create_socket(self):
+        """ Create new plain TCP socket. Hook that you may override. """
+        return socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+
     def configure_socket(self):
         """ Hook to configure socket parameters. """
         pass
 
+    def _create_openssl_connection(self):
+        """ Create new OpenSSL connection. Hook that you may override. """
+        return OpenSSL.SSL.Connection(self._certificate.get_context(), self._socket)
+
     def configure_connection(self):
         """ Hookt to configure SSL connection. """
         pass
 
-    def refresh(self):
-        """ Ensure socket is still alive. Reopen if needed. """
+    def _connect_and_handshake(self):
+        """ Connect to APNs and SSL handshake. Hook that you may override. """
+        self._connection.connect(self._address)
+        self._connection.do_handshake()
+
+    def _ensure_socket_open(self):
+        """ Refreshes socket. Hook that you may override. """
         if self._socket is None:
             try:
-                self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+                self._socket = self._create_socket()
                 self.configure_socket()
-
-                self._connection = OpenSSL.SSL.Connection(self._certificate.get_context(), self._socket)
+                self._connection = self._create_openssl_connection()
                 self.configure_connection()
-                self._connection.connect(self._address)
-                self._connection.do_handshake()
+                self._connect_and_handshake()
             except Exception:
                 self.close()
                 raise
 
+    def refresh(self):
+        """ Ensure socket is still alive. Reopen if needed. """
+        self._ensure_socket_open()
         self._readbuf = ""
         self._feedbackbuf = ""
         self._last_refresh = datetime.datetime.now()
@@ -685,7 +699,7 @@ class APNs(object):
 class Message(object):
     """ The notification message. """
 
-    def __init__(self, tokens, alert, badge=None, sound=None, expiry=None, **extra):
+    def __init__(self, tokens, alert=None, badge=None, sound=None, expiry=None, payload=None, **extra):
         """ The push notification to one or more device tokens.
 
             Read more `about payload
@@ -697,8 +711,18 @@ class Message(object):
                 - `badge` (int or str): badge number over the application icon or "
                 - `sound` (str): sound file to play on arrival.
                 - `expiry` (int or datetime or timedelta): timestamp when message will expire
+                - `payload` (dict): JSON-compatible dictionary with the
+                                    complete message payload. If supplied, it
+                                    is given instead of all the other, more
+                                    specific parameters.
                 - `extra` (kwargs): extra payload key-value pairs.
         """
+        if (payload is not None and (
+                alert is not None or badge is not None or sound is not None)):
+            # Raise an error if both `payload` and the more specific
+            # parameters are supplied.
+            raise ValueError("Payload specified together with alert/badge/sound.")
+
         if isinstance(tokens, basestring):
             tokens = [tokens]
 
@@ -706,6 +730,7 @@ class Message(object):
         self.alert = alert
         self.badge = badge
         self.sound = sound
+        self._payload = payload
         self.extra = extra
 
         if expiry is None:
@@ -769,8 +794,12 @@ class Message(object):
         """ List target device tokens. """
         return self._tokens
 
-    def batch(self, packet_size):
-        """ Returns binary serializer. """
+    @property
+    def payload(self):
+        """ Returns the payload content as a `dict`. """
+        if self._payload is not None:
+            return self._payload
+        
         aps = {
             # XXX: we do not check alert, which could be string or dict with extra options
             'alert': self.alert
@@ -789,7 +818,11 @@ class Message(object):
         if self.extra:
             ret.update(self.extra)
 
-        payload = json.dumps(ret)
+        return ret
+
+    def batch(self, packet_size):
+        """ Returns binary serializer. """
+        payload = json.dumps(self.payload)
         return Batch(self._tokens, payload, self.expiry, packet_size)
 
     def retry(self, failed_index, include_failed):
